#!/bin/bash

## CONSTANTS ########
TMP_PATH=/tmp
FILE_DATE=$(date +"%Y-%m-%d-%H-%M-%S")

QKVIEW_BIN_FILE=qkview-collect
QKVIEW_BIN_TAR_FILE=qkview-collect.tar
QKVIEW_BIN_MAIN_PATH=/usr/bin/$QKVIEW_BIN_FILE
QKVIEW_BIN_TMP_PATH=$TMP_PATH/tmp/$QKVIEW_BIN_FILE
QKVIEW_BIN_TAR_PATH=$TMP_PATH/$QKVIEW_BIN_TAR_FILE
QKVIEW_BIN_TAR_CMD="/bin/tar -xf $QKVIEW_BIN_TAR_PATH -C $TMP_PATH"

QKVIEW_TAR_FILE=qkview.tar.gz
QKVIEW_TAR_PATH=$TMP_PATH/$QKVIEW_TAR_FILE
QKVIEW_BASE_PATH=/usr/share/qkview
QKVIEW_CONFIG_PATH=/etc/qkview-collect/qkview-collect.config.yml

QKVIEW_BIN_PARAMS="-base-config-path $QKVIEW_BASE_PATH -overlay $QKVIEW_CONFIG_PATH -outfile $QKVIEW_TAR_PATH"

DIR_TMP=qkview_tmp_$FILE_DATE
DIR_OBJS="objs"
LOGFILE="qkview.log"
QKVIEW_CORE_FILE=corefiles.txt
QKVIEW_CORE_FILE_PATH=/shared/core

## VARIABLES ####################
Namespaces=""
K8sClient=""
CleanAfter=true
DebugLog=false
VerboseMode=false
PackagedExecution=false
DstDir="."
MaxFileSize=500
LocalTimeout=""
GlobalTimeout=""
PrivateKeys="false"
DssmSvcPort=26379

## HELP #############################

print_help() {
    echo ""
    echo "Qkview collector script."
    echo "   Run $QKVIEW_BIN_FILE for each container/pod for specified namespaces and tarball results into a single file"
    echo "Usage:"
    echo "   qkview-collectr.sh [parameters]"
    echo "Parameters:"
    echo "   -n   --namespaces      - list of namespaces, mandatory (example '-n=namespace1,namespace2')"
    echo "        --dir             - destination directory, default='$DstDir', example:'--dir=/tmp'"
    echo ""    
    echo "   -m   --max-file-size   - Maximum size of files to capture in Megabytes per container (qkview-collect -max-file-size )"
    echo "                            default: '$MaxFileSize', example: '-m=100'"
    echo "   -lt  --local-timeout   - Set the timeout for individual collection operations (seconds) (qkview-collect -local)"
    echo "                            ''=no timeout, default: '$LocalTimeout', example: '-lt=10'"
    echo "   -gt  --global-timeout  - Set the maximum collection timeout (seconds) per container (qkview-collect -global)"
    echo "                            ''=no timeout, default: '$GlobalTimeout', example: '-gt=60'"
    echo "   -pk  --private-keys    - Allow to collect private keys (flag), disabled by default"
    echo "                            (if not specified: qkview-collect -remove-privatekey-from-all-files)"
    echo ""
    echo "        --kubectl         - kubectl client (flag) (default)"
    echo "        --oc              - use oc client (flag)"
    echo "   -dp  --dssm-port       - port for f5-dssm-sentinel service (default 26379), example '-dp=26380'"
    echo ""
    echo "   -d   --debug           - enables debug logs (flag)"    
    echo "   -v   --verbose         - enables verbose mode (flag)"
    echo "   -h   --help            - this help (flag)"
    echo "   -nc  --noclean         - disable removing tmp directory after script (flag)"
    echo ""
    echo "Example:"
    echo "    qkview-collectr.sh -n=namespace1,namespace2 -d -v --dir=/tmp"
    echo ""
}

## LOGGING ######################

function log() {

    local Level=$1
    local Msg=$2
    local WriteLogToFile=$3

    msg="$(date +"%Y/%m/%d %H:%M:%S:%N")\t$Level:\t$Msg"
    if ([[ "$Level" != "DEBUG" ]]) || ([[ "$DebugLog" = true ]]); then
        echo -e "$msg"
    fi

    if ([[ "$#" = 2 ]]) || ([[ $# > 2 ]] && [[ "$WriteLogToFile" = true ]]); then
        if [[ "$LOGFILE" != "" ]]; then
            echo -e "$msg" >> $DstDir/$DIR_TMP/$LOGFILE
        fi
    fi

}

function logDebug() {

    local Level="DEBUG"
    local Msg=$1
    local WriteLogToFile=$2
    log $Level "$Msg" $WriteLogToFile

}

function logInfo() {

    local Level="INFO"
    local Msg=$1
    local WriteLogToFile=$2
    log $Level "$Msg" $WriteLogToFile

}

function logError() {

    local Level="ERROR"
    local Msg=$1
    local WriteLogToFile=$2
    log $Level "$Msg" $WriteLogToFile

}

## CONFIG ###########################

print_var() {

    logDebug "Variables:"
    logDebug "  Namespaces=$Namespaces"
    logDebug "  K8sClient=$K8sClient"
    logDebug "  CleanAfter=$CleanAfter"
    logDebug "  DebugLog=$DebugLog"
    logDebug "  VerboseMode=$VerboseMode"
    logDebug "  DstDir=$DstDir"
    logDebug "  MaxFileSize=$MaxFileSize"
    logDebug "  LocalTimeout=$LocalTimeout"
    logDebug "  GlobalTimeout=$GlobalTimeout"
    logDebug "  PrivateKeys=$PrivateKeys"
    logDebug "  DssmSvcPort=$DssmSvcPort"

}

parse_argument() {

    logDebug "Parsing argument '$1'" false
    case $1 in
        -n=*|--namespaces=*)
            Namespaces="${1#*=}"
            ;;
   
        --kubectl)
            if [[ "$K8sClient" = "oc" ]]; then
                logError "Incompatible options is specified: --kubectl/--oc" false
                exit 1
            fi
            K8sClient="kubectl"
            ;;
        --oc)
            if [[ "$K8sClient" = "kubectl" ]]; then
                logError "Incompatible options is specified: --kubectl/--oc" false
                exit 1
            fi
            K8sClient="oc"
            ;;

        --dir=*)
            DstDir="${1#*=}"
            ;;
        -m=*|--max-file-size=*)
            MaxFileSize="${1#*=}"
            ;;
        -lt=*|--local-timeout=*)
            LocalTimeout="${1#*=}"
            ;;
        -gt=*|--global-timeout=*)
            GlobalTimeout="${1#*=}"
            ;;
        -pk|--private-keys)
            PrivateKeys="true"
            ;;
        -dp=*|--dssm-port=*)
            DssmSvcPort="${1#*=}"
            ;;

        -d|--debug)
            DebugLog=true
            ;;
        -v|--verbose)
            VerboseMode=true
            ;;
        -h|--help)
            print_help
            exit 0
            ;;
        -nc|--noclean)
            CleanAfter=false
            ;;
        --packaged-execution)
            PackagedExecution=true
            ;;
        *)
            logError "No handler for argument '$1'" false
            exit 1
            ;;
    esac

}

init() {

    local WriteLogToFile=false

    # Creating tmp dir
    logInfo "Creating temporary directory for collected data: $DstDir/$DIR_TMP" $WriteLogToFile
    local_exec "mkdir -p $DstDir/$DIR_TMP" $WriteLogToFile
    if [ $? -ne 0 ]; then
        logError "Error on creating dir $DstDir/$DIR_TMP" $WriteLogToFile
        exit 1
    fi

}

k8s_client_selector() {

    local WriteLogToFile=false

    if [[ "$K8sClient" != "" ]]; then
        logDebug "Client is specified: $K8sClient" $WriteLogToFile
        local_exec "$K8sClient version" $WriteLogToFile
        if [ $? -ne 0 ]; then
            logError "$K8sClient client not found" $WriteLogToFile
            exit 1
        fi
    else
        logDebug "Client is not specified. Detecting..." $WriteLogToFile
    fi

    # kubectl client
    if [[ "$K8sClient" = "" ]]; then
        K8sClient="kubectl"
        local_exec "$K8sClient version" $WriteLogToFile
        if [ $? -ne 0 ]; then
            logError "$K8sClient client not found" $WriteLogToFile
            K8sClient=""
        fi
    fi

    # oc client
    if [[ "$K8sClient" = "" ]]; then
        K8sClient="oc"
        local_exec "$K8sClient version" $WriteLogToFile
        if [ $? -ne 0 ]; then
            logError "$K8sClient client not found" $WriteLogToFile
            K8sClient=""
        fi
    fi

    # default value
    if [[ "$K8sClient" = "" ]]; then
        exit 1
    fi

}

get_namespaces() {

    local WriteLogToFile=false
    local NamespaceItem

    if [[ "$Namespaces" != "" ]]; then
        logDebug "Namespaces is specified: $Namespaces" $WriteLogToFile
    else
        logError "Error: Namespace is not specified" $WriteLogToFile
        exit 1
    fi

    # Check namespaces
    local NamespaceList=($(echo "$Namespaces" | tr "," " "))
    for NamespaceItem in "${NamespaceList[@]}"
    do
        local_exec "$K8sClient get namespace $NamespaceItem" $WriteLogToFile true
        if [ $? -ne 0 ]; then
            logError "Error: Namespace $NamespaceItem does not exist" $WriteLogToFile
            exit 1
        fi
    done

}

## CLIENT FUNCTIONS ########################

# local_exec
#  
# $1 - command
# $2 - put log to file (true/false)
# $3 - do not redirect stdout to null (true/false)
local_exec() {

    local Command=$1
    local WriteLogToFile=$2
    local StdoutRedirect=$3

    if [[ "$VerboseMode" = "true" ]]; then
        cmd="$Command"
    else
        if [[ "$StdoutRedirect" = "false" ]]; then
            cmd="$Command 2>/dev/null"
        else
            cmd="$Command 1>/dev/null 2>/dev/null"
        fi
    fi

    logDebug "  Command: $Command" $WriteLogToFile
    eval "$cmd"
    retVal=$?
    logDebug "    Result code: $retVal" $WriteLogToFile

    return $retVal

}

# k8s_exec_run
#  
# $1 - command (like 'cp', 'logs', 'get pods')
# $2 - namespcace
# $3 - pod
# $4 - additional command parameters (e.g. '-c <container> -- bash')
# $5 - do not redirect stdout to null (true/false)
k8s_exec_run() {

    local Verb=$1
    local Namespace=$2
    local Pod=$3
    local AddCmd=$4
    local StdoutRedirect=$5
    local WriteLogToFile=true
    local Cmd="$K8sClient"
   
    Cmd="$Cmd $Verb"

    if [[ "$Namespace" != "" ]]; then
        Cmd="$Cmd -n $Namespace"
    fi

    if [[ "$Pod" != "" ]]; then
        Cmd="$Cmd $Pod"
    fi
    
    if [[ "$AddCmd" != "" ]]; then
        Cmd="$Cmd $AddCmd"
    fi

    local_exec "$Cmd" $WriteLogToFile $StdoutRedirect
    return $?

}

# k8s_exec
#  
# $1 - namespcace
# $2 - pod
# $3 - container
# $4 - command
k8s_exec() {

    local Namespace=$1
    local Pod=$2
    local Container=$3
    local Command=$4
    k8s_exec_run "exec -i" "$Namespace" "$Pod" "-c $Container -- $Command"
    return $?

}

# k8s_cp
#  
# $1 - namespcace
# $2 - src file
# $3 - dst file
# $4 - container
k8s_cp() {

    local Namespace=$1
    local Src=$2
    local Dst=$3
    local Container=$4
    k8s_exec_run "cp" "$Namespace" "$Src $Dst" "-c $Container"
    return $?

}

## COLLECTOR FUNCTIONS ########################

handle_cluster() {
    local NamespaceItem
    local Namespace=""

    logInfo "Handle namespaces: $Namespaces"

    # Dssm service
    DssmSrvcs=()
    
    local NamespaceList=($(echo "$Namespaces" | tr "," " "))
    for NamespaceItem in "${NamespaceList[@]}"
    do
        handle_namespace $NamespaceItem
    done

    # Collecting mrfdb output
    handle_dssm

    # Collecting cluster level objects
    local_exec "mkdir -p $DstDir/$DIR_TMP/$DIR_OBJS"
    handle_kinds $Namespace

    logInfo "Collecting cluster info"
    k8s_exec_run "cluster-info dump" "" "" "> $DstDir/$DIR_TMP/cluster-info" false

}

handle_namespace() {

    local Namespace=$1
    logInfo "+ Collecting namespace: $Namespace"
    
    local_exec "mkdir -p $DstDir/$DIR_TMP/$Namespace/"

    PodList=($($K8sClient -n $Namespace get pods --template '{{range .items}}{{.metadata.name}} {{end}}'))
    for PodItem in "${PodList[@]}"
    do
        handle_pod $Namespace $PodItem
    done

    # Collecting objects
    local_exec "mkdir -p $DstDir/$DIR_TMP/$Namespace/$DIR_OBJS"
    handle_kinds $Namespace
    
    logInfo "- Collecting namespace: $Namespace is done"

}

handle_kinds() {

    local Namespace=$1
    local SkipItem=false
    local KindList
    local ObjList

    if [[ "$Namespace" != "" ]]; then 
        KindList=($($K8sClient api-resources -o=name --namespaced=true --sort-by=name | tr "\n" " "))
    else
        KindList=($($K8sClient api-resources -o=name --namespaced=false --sort-by=name | tr "\n" " "))
    fi

    for KindItem in "${KindList[@]}"
    do
        if [[ "$Namespace" != "" ]]; then
            ObjList=($($K8sClient -n $Namespace get $KindItem --template '{{range .items}}{{.metadata.name}} {{end}}'))
        else
            ObjList=($($K8sClient get $KindItem --template '{{range .items}}{{.metadata.name}} {{end}}'))
        fi

        if [[ ${#ObjList[@]} > 0 ]]; then
            handle_kind "$Namespace" "$KindItem"
        fi
    done

}

handle_kind() {

    local Namespace=$1
    local Kind=$2
    local KindPath=$Kind
    local DirPath=$DIR_OBJS

    if [[ "$Namespace" != "" ]]; then
        KindPath="$Namespace/$KindPath"
        DirPath="$Namespace/$DirPath"
    fi
    DirPath="$DstDir/$DIR_TMP/$DirPath"

    logInfo "Collecting kind $KindPath"
    handle_objs "$Namespace" "$Kind" "" "$DirPath/$Kind"

}

handle_objs() {

    local Namespace=$1
    local Kind=$2
    local ObjName=$3
    local Path=$4

    k8s_exec_run "get $Kind" "$Namespace" "$ObjName" "-o yaml > $Path.yaml" false
    if [ $? -ne 0 ]; then
        logError "Error on collecting $Kind/$ObjName"
    fi

    k8s_exec_run "describe $Kind" "$Namespace" "$ObjName" "> $Path.describe" false
    if [ $? -ne 0 ]; then
        logError "Error on collecting describe $Kind/$ObjName"
    fi

    # Get DSSM srvc
    if [[ $Kind == "pod" ]]; then
        DssmSrvc=($(cat $Path.describe | grep 'SESSIONDB_EXTERNAL_SERVICE:' | sed 's/[ \t]*SESSIONDB_EXTERNAL_SERVICE:[ \t]*//g'))
        if [[ $DssmSrvc != "" ]]; then
            logDebug "Found DSSM service in Pod env variables: $DssmSrvc"
            if [[ ! "$DssmSrvcs" =~ ^$DssmSrvc,* ]]; then
                logDebug "DSSM service data will be collected: $DssmSrvc"
                DssmSrvcs+=("$DssmSrvc,$Namespace,$ObjName")
            fi
        fi
    fi
}

handle_dssm() {

    local Cmd
    for DssmItem in ${DssmSrvcs[@]}; do
        IFS="," read -a DssmSrvc <<< "$DssmItem"
        logInfo "Collecting mrfdb output for ${DssmSrvc[0]} via ${DssmSrvc[1]}/${DssmSrvc[2]}"
        Cmd="mrfdb -ipport=${DssmSrvc[0]}:$DssmSvcPort -serverName=server -displayAllBins"
        k8s_exec_run "exec -i" "${DssmSrvc[1]}" "${DssmSrvc[2]}" "-c debug -- $Cmd > $DstDir/$DIR_TMP/mrfdb_${DssmSrvc[0]}.out" false
        if [ $? -ne 0 ]; then
            logError "Error during collecting DSSM data for ${DssmSrvc[0]} via pod ${DssmSrvc[1]}/${DssmSrvc[2]}"
        fi
    done

}

handle_pod() {

    local Namespace=$1
    local Pod=$2

    logInfo "++ Collecting pod: $Namespace/$Pod"
    
    local_exec "mkdir -p $DstDir/$DIR_TMP/$Namespace/$Pod/"

    ContainerList=($($K8sClient -n $Namespace get pod $Pod -o jsonpath={.spec.containers[*].name}))
    for ContainerItem in "${ContainerList[@]}"
    do
        handle_container $Namespace $Pod $ContainerItem
    done

    # Collecting pod description
    handle_objs "$Namespace" "pod" "$Pod" "$DstDir/$DIR_TMP/$Namespace/$Pod/$Pod"

    logInfo "-- Collecting pod: $Namespace/$Pod is done"
}

handle_container() {
    
    local Namespace=$1
    local Pod=$2
    local Container=$3
    
    logInfo "+++ Collecting container: $Namespace/$Pod/$Container"

    local_exec "mkdir -p $DstDir/$DIR_TMP/$Namespace/$Pod/$Container/"

    local QkviewBinPath=""
    local CopyQkviewTar=false
    local CleanQkviewTar=false

    # Check for core files in /shared/core
    coreCmd="ls -1 $QKVIEW_CORE_FILE_PATH | grep -v coretmmpipe"

    k8s_exec $Namespace $Pod $ContainerItem $coreCmd
    if [ $? -ne 0 ]; then
        logError "QkView core files not found in $QKVIEW_CORE_FILE_PATH"
    else
        logDebug "QkView core files found in $QKVIEW_CORE_FILE_PATH"
        k8s_exec_run "exec -i" "$Namespace" "$Pod" "-c $ContainerItem -- $coreCmd > $DstDir/$DIR_TMP/$Namespace/$Pod/$Container/$QKVIEW_CORE_FILE" false
    fi

    # Check qkvew-collect binary in /usr/bin/
    k8s_exec $Namespace $Pod $ContainerItem "ls -l $QKVIEW_BIN_MAIN_PATH"
    if [ $? -ne 0 ]; then
        logError "QkView binary not found in $QKVIEW_BIN_MAIN_PATH"
        QkviewBinPath=""
    else
        logDebug "QkView binary found in $QKVIEW_BIN_MAIN_PATH"
        QkviewBinPath=$QKVIEW_BIN_MAIN_PATH
    fi
    
    # Additional qkvew-collect params
    QKVIEW_BIN_ADD_PARAMS=""
    if [[ "$PrivateKeys" == false ]]; then
        QKVIEW_BIN_ADD_PARAMS="$QKVIEW_BIN_ADD_PARAMS -remove-privatekey-from-all-files"
    fi
    if [[ "$LocalTimeout" != "" ]]; then
        QKVIEW_BIN_ADD_PARAMS="$QKVIEW_BIN_ADD_PARAMS -local $LocalTimeout"
    fi
    if [[ "$GlobalTimeout" != "" ]]; then
        QKVIEW_BIN_ADD_PARAMS="$QKVIEW_BIN_ADD_PARAMS -global $GlobalTimeout"
    fi
    QKVIEW_BIN_ADD_PARAMS="$QKVIEW_BIN_ADD_PARAMS -max-file-size $MaxFileSize"

    # Run qkview-collect
    if [[ "$QkviewBinPath" != "" ]]; then
        logInfo "Run qkview collector"
        k8s_exec $Namespace $2 $ContainerItem "$QkviewBinPath $QKVIEW_BIN_PARAMS$QKVIEW_BIN_ADD_PARAMS"
        if [ $? -ne 0 ]; then
            logError "Qkview run error"
        else
            CopyQkviewTar=true
        fi
    fi

    # Copy tar to local path
    if [[ "$CopyQkviewTar" == true ]]; then
        logDebug "Copying qkview collected data tar to local client"
        k8s_cp $Namespace $Pod:$QKVIEW_TAR_PATH $DstDir/$DIR_TMP/$Namespace/$Pod/$Container/$QKVIEW_TAR_FILE $Container
        if [ $? -ne 0 ]; then
            logError "Qkview copy collected data tar error"
        else
            CleanQkviewTar=true
        fi
    fi

    # Clean tar on container
    if [[ "$CleanQkviewTar" == true ]]; then
        logDebug "Clean qkview collect data tar on container"
        k8s_exec $Namespace $Pod $ContainerItem "rm -f $QKVIEW_TAR_PATH"
        if [ $? -ne 0 ]; then
            logError "Qkview clean collected data tar error"
        fi
    fi

    # Collecting container logs
    k8s_exec_run "logs" $Namespace $Pod "-c $ContainerItem > $DstDir/$DIR_TMP/$Namespace/$Pod/$Container/$Container.log" false
    if [ $? -ne 0 ]; then
        logError "Error on collecting logs for $Namespace/$Pod/$ContainerItem"
    fi
    
    logInfo "--- Collecting container: $Namespace/$Pod/$Container is done"

}

package_results() {

    logInfo "Creating final tarball"
    if [[ "$PackagedExecution" == true ]]; then
        local FileName="qkview-packaging.tar.gz"
    else
        local FileName="qkview-$FILE_DATE.tar.gz"
    fi

    local_exec "tar -zcvf $DstDir/$FileName -C $DstDir/$DIR_TMP ."
    if [ $? -ne 0 ]; then
        logError "Tarball creating error"
    else
        logInfo "Tarball created: $DstDir/$FileName"
    fi

}

clean_results() {

    if [[ "$CleanAfter" != true ]]; then
        return
    fi
    logDebug "Clean temporary directory"
    
    local_exec "rm -rf $DstDir/$DIR_TMP" false
    if [ $? -ne 0 ]; then
        logError "Clean temporary directory error" false
    fi

}

## MAIN CODE ########################

# Parse arguments
for i in "$@"; do
    parse_argument $i
done

# Get config 
k8s_client_selector
get_namespaces

# Initialization
init
print_var

# Collect data
handle_cluster
package_results

# Cleaning results
clean_results
